/*=============================================================================*/
/*
	file:
	author:
	date:
	notes:
*/
/*=============================================================================*/

/*=============================================================================*/
/*
 		Includes
*/
/*=============================================================================*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "MediaPlayer.h"
#include "FrameTimer.h"

/*
 *	This is used to allow programs to be compiled both on windows and irix
 */
#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#include <windows.h>
	#include <GL/glut.h>
#else
	#include "glut.h"
#endif


/*=============================================================================*/
/*
 		Function Prototypes

 		A quick program breakdown....
*/
/*=============================================================================*/


void InitGL(void);					/*	sets initial openGL settings, load any data etc */

void Display(void);					/*	called whenever screen needs to be re-drawn */

void ReshapeWindow(int,int);		/*	called whenever the window needs to be resized */

void KeyDown(unsigned char,int,int);/*	called when a key is pressed */

void SpecialKeyDown(int,int,int);	/*	called when a special key is pressed */

void SpecialKeyUp(int,int,int);		/*	called when a special key is released */

void Idle(void);					/*	called when the app has nothing to do */


/*=============================================================================*/
/*
 		Some Useful macros and constants

 		Some of these may be of some use...
*/
/*=============================================================================*/

/*
 *	Numeric PI constants
 */
#define PI 3.1415926535897932384626433832795
#define PI_OVER_2 1.5707963267948966192313216916398
#define PI_OVER_180 0.017453292519943295769236907684886
#define _180_OVER_PI 57.295779513082320876798154814105

/*
 *	degree and radian conversions
 */
#define deg_to_rad(x) ((x)*PI_OVER_180)
#define rad_to_deg(x) ((x)*_180_OVER_PI)

/*
 *	A wrapper around the string compare function for readability
 */
#define AreStringsEqual(X,Y) (strcmp(X,Y)==0)



/*=============================================================================*/
/*
 		Constants

 		use these constants to set some default settings for your app
*/
/*=============================================================================*/

#define APP_TITLE "Glut Template App"

#define DEFAULT_WINDOW_WIDTH  640
#define DEFAULT_WINDOW_HEIGHT 480
#define DEFAULT_WINDOW_POS_X  100
#define DEFAULT_WINDOW_POS_Y  100
#define DEFAULT_CAMERA_ANGLE  45.0f
#define DEFAULT_CAMERA_ASPECT 1.333f
#define DEFAULT_CAMERA_NEAR   0.1f
#define DEFAULT_CAMERA_FAR    100.0f


/*=============================================================================*/
/*
 		Global Variables

 		query the following variables in your app to find out the window dimensions

 		g_iWidth	-	window width
 		g_iHeight	-	window height
 		g_iPosX		-	window position x
 		g_iPosY		-	window position y
*/
/*=============================================================================*/

/*	This global variable will hold the current window width	*/
int g_iWidth = DEFAULT_WINDOW_WIDTH;

/*	This global variable will hold the current window height	*/
int g_iHeight = DEFAULT_WINDOW_HEIGHT;

/*	This global variable will hold the current window width	*/
int g_iPosX = DEFAULT_WINDOW_POS_X;

/*	This global variable will hold the current window height	*/
int g_iPosY = DEFAULT_WINDOW_POS_Y;




/*=============================================================================*/
/*
		3D Sound stuff is here
*/
/*=============================================================================*/

/*
 *	An array of the current special-key states
 */
int Keys[128] = {0};

/*
 *	The positions of the sounds in 3D space
 */
float Position1[] = { 10,2,-7 };
float Position2[] = { -8,2,-7 };
float Position3[] = {  1,2,1  };
float Position4[] = { -2,2,9  };
float Position5[] = {  9,2,5  };

/*
 *	An array of identifier that reference our sounds (a bit like texture objects in OGL)
 */
unsigned int g_aSounds[5];

/*
 *	The up direction of the listener
 */
float up[] = {0,1,0};

/*
 *	Some other bits of data about the listener
 */
float ListenerAngle = 0.0f;
float ListenerDirection[3] = {0,0,-1};
float ListenerPosition [3] = {0,0,0};
float ListenerVelocity [3] = {0,0,0};

/*
 *	Updates all the sound info in the scene.
 */
void ListenerSetParams()
{
	/*
	 *	Update position and orientation of listener
	 */
	MediaSetPosition( ListenerPosition );
	MediaSetOrientation( ListenerDirection,up );

	/*
	 *	Calculate and update the velocity
	 */
	ListenerVelocity[0] = 3 * ListenerDirection[0];
	ListenerVelocity[2] = 3 * ListenerDirection[2];

	MediaSetVelocity( ListenerVelocity );

	/*
	 *	Probably don't need to do this each frame...
	 */
	MediaSetSoundPosition(g_aSounds[0],Position1);
	MediaSetSoundPosition(g_aSounds[1],Position2);
	MediaSetSoundPosition(g_aSounds[2],Position3);
	MediaSetSoundPosition(g_aSounds[3],Position4);
	MediaSetSoundPosition(g_aSounds[4],Position5);

	/*
	 *	Ensure that all of our sound changes have been actioned before rendering
	 */
	MediaUpdate3D();
}

void ListenerWalkForward()
{
	ListenerPosition[0] -= 5.0f * GetFrameTime() * ListenerDirection[0];
	ListenerPosition[2] -= 5.0f * GetFrameTime() * ListenerDirection[2];
}

void ListenerWalkBackward()
{
	ListenerPosition[0] += 5.0f * GetFrameTime() * ListenerDirection[0];
	ListenerPosition[2] += 5.0f * GetFrameTime() * ListenerDirection[2];
}

void ListenerRotateLeft()
{
	ListenerAngle -= 180.0f * GetFrameTime();
}

void ListenerRotateRight()
{
	ListenerAngle += 180.0f * GetFrameTime();
}

/*
 *	This function draws boxes where the sound appears from
 */
void RenderSoundPositions()
{
	glPushMatrix();
		glColor3f(1,0,0);
		glTranslatef(Position1[0],Position1[1],Position1[2]);
		glutSolidCube(0.5f);
	glPopMatrix();

	glPushMatrix();
		glColor3f(0,1,0);
		glTranslatef(Position2[0],Position2[1],Position2[2]);
		glutSolidCube(0.5f);
	glPopMatrix();

	glPushMatrix();
		glColor3f(0,0,1);
		glTranslatef(Position3[0],Position3[1],Position3[2]);
		glutSolidCube(0.5f);
	glPopMatrix();

	glPushMatrix();
		glColor3f(1,1,0);
		glTranslatef(Position4[0],Position4[1],Position4[2]);
		glutSolidCube(0.5f);
	glPopMatrix();

	glPushMatrix();
		glColor3f(1,0,1);
		glTranslatef(Position5[0],Position5[1],Position5[2]);
		glutSolidCube(0.5f);
	glPopMatrix();
}

/*
 *	This function updates the user input and listener location
 */
void ListenerCheckUserInput()
{
	/*
	 *	Check rotation
	 */
	if (Keys[ GLUT_KEY_LEFT ]) {
		ListenerRotateLeft();
	}
	if (Keys[ GLUT_KEY_RIGHT ]) {
		ListenerRotateRight();
	}

	/*
	 *	Update Listener direction vector
	 */
	ListenerDirection[0] = -sin( deg_to_rad( ListenerAngle ) );
	ListenerDirection[2] = cos( deg_to_rad( ListenerAngle ) );

	/*
	 *	Check movement
	 */
	if (Keys[ GLUT_KEY_UP ]) {
		ListenerWalkForward();
	}
	if (Keys[ GLUT_KEY_DOWN ]) {
		ListenerWalkBackward();
	}

	/*
	 *	Set the Listener and sound params
	 */
	ListenerSetParams();
}






/*=============================================================================*/
/*
		Custom Application Initialisation

		extend these functions to process additional command line arguments or
		to add extra openGL initialisation code.
*/
/*=============================================================================*/


/*
 *	This function is called after the openGL window has been created.
 *	It allows you to set up openGL for your application.
 */
void InitGL()
{
	/*
	 *	Set the background colour to grey
	 */
	glClearColor(0.5f,0.5f,0.5f,0.0f);

	MediaInit();

	/*
	 *	Load all of the sounds
	 */
	g_aSounds[0] = MediaLoad("Media/1.wav");
	g_aSounds[1] = MediaLoad("Media/2.wav");
	g_aSounds[2] = MediaLoad("Media/3.wav");
	g_aSounds[3] = MediaLoad("Media/4.wav");
	g_aSounds[4] = MediaLoad("Media/5.wav");


	/*
	 *	Set the max distance of each sound. If you are over 20
	 *	meters from the sound, it will not be heard.
	 */
	MediaSetMaxDistance( g_aSounds[0], 20 );
	MediaSetMaxDistance( g_aSounds[1], 20 );
	MediaSetMaxDistance( g_aSounds[2], 20 );
	MediaSetMaxDistance( g_aSounds[3], 20 );
	MediaSetMaxDistance( g_aSounds[4], 20 );

	MediaSetMinDistance( g_aSounds[0], 1 );
	MediaSetMinDistance( g_aSounds[1], 1 );
	MediaSetMinDistance( g_aSounds[2], 1 );
	MediaSetMinDistance( g_aSounds[3], 1 );
	MediaSetMinDistance( g_aSounds[4], 1 );

}



/*=============================================================================*/
/*
		Glut Callback Functions (Window Callbacks)

		These callbcaks deal with events relating to the window.

		Display			-	Called when the window needs to be re-drawn
		ReshapeWindow	-	Called when the window is resized
*/
/*=============================================================================*/


/*
 *	This function is called whenever the window needs to be re-drawn
 */
void Display(void)
{
	/*
	 *	Clear the screen and the depth buffer
	 */
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	/*
	 *	Load the identity transformation matrix to clear all previous
	 *	transformations.
	 */
	glLoadIdentity();

	glRotatef(ListenerAngle,0,1,0);
	glTranslatef(-ListenerPosition[0],-1.25,-ListenerPosition[2]);

	glColor3f(0,0,0);
	glBegin(GL_LINES);
	for(int i=-10;i<=10;++i)
	{
		glVertex3f(i,0,10);
		glVertex3f(i,0,-10);
		glVertex3f(10,0,i);
		glVertex3f(-10,0,i);
	}
	glEnd();



	glPushMatrix();

	/*
	 *	Drawing code goes here ...
	 */
	RenderSoundPositions();

	glPopMatrix();

	/*
	 *	Swap the buffers over to see what we just drew...
	 */
	glutSwapBuffers();
}

/*
 *	This function is called whenever the size of the window changes
 *	w	=	the new width of the window
 *	h	=	the new height of the window
 */
void ReshapeWindow(int w,int h)
{
	/*
	 *	Store new width & height
	 */
	g_iWidth  = w;
	g_iHeight = h;

	/*
	 *	get new window position
	 */
	g_iPosX = glutGet(GLUT_WINDOW_X);
	g_iPosY = glutGet(GLUT_WINDOW_Y);

	/*
	 *	Specify the drawable region of the screen
	 */
	glViewport(0,0,w,h);

	/*
	 *	Set the matrix mode to GL_PROJECTION so that we can specify
	 *	the 3D to 2D projection matrix
	 */
	glMatrixMode(GL_PROJECTION);

	/*
	 *	Load the identity matrix
	 */
	glLoadIdentity();

	/*
	 *	Set a perspective projection
	 */
	gluPerspective( DEFAULT_CAMERA_ANGLE  ,
					DEFAULT_CAMERA_ASPECT ,
					DEFAULT_CAMERA_NEAR   ,
					DEFAULT_CAMERA_FAR     );

	/*
	 *	Go back to model mode..
	 */
	glMatrixMode(GL_MODELVIEW);
}


/*=============================================================================*/
/*
		Glut Callback Functions (Idle Processing)

		Idle			-	Called when the app has nothing to do
*/
/*=============================================================================*/


/*
 *	This function is called whenever the app has nothing to do
 */
void Idle(void)
{

	UpdateFrameTime ();
	ListenerCheckUserInput();

	/*
	 *	Force a re-draw of the screen
	 */
	glutPostRedisplay();
}


/*=============================================================================*/
/*
		Glut Callback Functions (User Input Callbacks)

		These callbcaks deal with events relating to user input events

		KeyDown			-	Called when a normal key is pressed
		KeyUp			-	Called when a normal key is released
		SpecialKeyDown	-	Called when a function or cursor key is pressed
		SpecialKeyUp	-	Called when a function or cursor key is released
		MouseButton		-	Called when a mouse button is pressed or released
		MouseMove		-	Called when the mouse moves and the button is pressed
*/
/*=============================================================================*/


/*
 *	This function handles the keyboard interupts generated when a key is pressed down
 *
 *	key		=	the ascii key code, ie 'a'
 *	x		=	x-coord of mouse pointer when the key-press occurred
 *	y		=	y-coord of mouse pointer when the key-press occurred
 */
void KeyDown(unsigned char key,int x, int y)
{
	/*
	 *	If key == ESC, exit
	 */

	switch(key)
	{
	case 27:
		exit(0);
	case '1':
		MediaPlay(g_aSounds[0]);
		break;
	case '2':
		MediaPlay(g_aSounds[1]);
		break;
	case '3':
		MediaPlay(g_aSounds[2]);
		break;
	case '4':
		MediaPlay(g_aSounds[3]);
		break;
	case '5':
		MediaPlay(g_aSounds[4]);
		break;
	default:
		break;
	}
}
/*
 *	This function handles the keyboard interupts generated when a special key is pressed down
 *
 *	key		=	the special key code, ie GLUT_KEY_F1, GLUT_KEY_UP etc...
 *	x		=	x-coord of mouse pointer when the key-press occurred
 *	y		=	y-coord of mouse pointer when the key-press occurred
 */
void SpecialKeyDown(int key,int x, int y)
{
	Keys[key]=1;
}

/*
 *	This function handles the keyboard interupts generated when a special key is released
 *
 *	key		=	the special key code, ie GLUT_KEY_F1, GLUT_KEY_UP etc...
 *	x		=	x-coord of mouse pointer when the key-release occurred
 *	y		=	y-coord of mouse pointer when the key-release occurred
 */
void SpecialKeyUp(int key,int x, int y)
{
	Keys[key]=0;
}


/*=============================================================================*/
/*
		Main Function
*/
/*=============================================================================*/

/*
 *	The main function...
 *
 *	argc	-	the number of command line arguments
 *	argv	-	an array of command line strings
 *
 */
void main(int argc,char *argv[])
{

	/*
	 *	Initialise glut
	 */
	glutInit(&argc,argv);

	/*
	 *	Set the display mode to support RGBA, Depth & double buffering
	 */
	glutInitDisplayMode(GLUT_DEPTH|
						GLUT_RGB |
						GLUT_DOUBLE);

	/*
	 *	Set the size of the window
	 */
	glutInitWindowSize( g_iWidth,
						g_iHeight );

	/*
	 *	Set the position of the window
	 */
	glutInitWindowPosition( g_iPosX ,
							g_iPosY );


	/*
	 *	Create the window
	 */
	glutCreateWindow(APP_TITLE);

	/*
	 *	Call the custom openGL initialisation
	 */
	InitGL();

	/*
	 *	Set all of the glut callback functions
	 */
	glutDisplayFunc(Display);
	glutReshapeFunc(ReshapeWindow);
	glutKeyboardFunc(KeyDown);
	glutSpecialFunc(SpecialKeyDown);
	glutSpecialUpFunc(SpecialKeyUp);
	glutIdleFunc(Idle);

	InitFrameTime();

	/*
	 *	Enter the main program loop
	 */
	glutMainLoop();

}
